# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: app/app_payload.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.app import app_definition_pb2 as app_dot_app__definition__pb2
from union.internal.common import identifier_pb2 as common_dot_identifier__pb2
from union.internal.common import list_pb2 as common_dot_list__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61pp/app_payload.proto\x12\x0c\x63loudidl.app\x1a\x18\x61pp/app_definition.proto\x1a\x17\x63ommon/identifier.proto\x1a\x11\x63ommon/list.proto\x1a\x17validate/validate.proto\">\n\rCreateRequest\x12-\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x03\x61pp\"5\n\x0e\x43reateResponse\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\"\x99\x01\n\nGetRequest\x12;\n\x06\x61pp_id\x18\x01 \x01(\x0b\x32\x18.cloudidl.app.IdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x05\x61ppId\x12;\n\x07ingress\x18\x02 \x01(\x0b\x32\x15.cloudidl.app.IngressB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x07ingressB\x11\n\nidentifier\x12\x03\xf8\x42\x01\"2\n\x0bGetResponse\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\">\n\rUpdateRequest\x12-\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x03\x61pp\"5\n\x0eUpdateResponse\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\"J\n\rDeleteRequest\x12\x39\n\x06\x61pp_id\x18\x01 \x01(\x0b\x32\x18.cloudidl.app.IdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x05\x61ppId\"\x10\n\x0e\x44\x65leteResponse\"\x8d\x02\n\x0bListRequest\x12\x36\n\x07request\x18\x01 \x01(\x0b\x32\x1c.cloudidl.common.ListRequestR\x07request\x12\x1b\n\x03org\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01H\x00R\x03org\x12M\n\ncluster_id\x18\x03 \x01(\x0b\x32\".cloudidl.common.ClusterIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\tclusterId\x12H\n\x07project\x18\x04 \x01(\x0b\x32\".cloudidl.common.ProjectIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x07projectB\x10\n\tfilter_by\x12\x03\xf8\x42\x01\"K\n\x0cListResponse\x12%\n\x04\x61pps\x18\x01 \x03(\x0b\x32\x11.cloudidl.app.AppR\x04\x61pps\x12\x14\n\x05token\x18\x02 \x01(\tR\x05token\"\x90\x02\n\x0cWatchRequest\x12\x1b\n\x03org\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01H\x00R\x03org\x12M\n\ncluster_id\x18\x02 \x01(\x0b\x32\".cloudidl.common.ClusterIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\tclusterId\x12H\n\x07project\x18\x03 \x01(\x0b\x32\".cloudidl.common.ProjectIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x07project\x12;\n\x06\x61pp_id\x18\x04 \x01(\x0b\x32\x18.cloudidl.app.IdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x05\x61ppIdB\r\n\x06target\x12\x03\xf8\x42\x01\"2\n\x0b\x43reateEvent\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\"m\n\x0bUpdateEvent\x12\x32\n\x0bupdated_app\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\nupdatedApp\x12*\n\x07old_app\x18\x02 \x01(\x0b\x32\x11.cloudidl.app.AppR\x06oldApp\"2\n\x0b\x44\x65leteEvent\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\"\xd8\x01\n\rWatchResponse\x12>\n\x0c\x63reate_event\x18\x01 \x01(\x0b\x32\x19.cloudidl.app.CreateEventH\x00R\x0b\x63reateEvent\x12>\n\x0cupdate_event\x18\x02 \x01(\x0b\x32\x19.cloudidl.app.UpdateEventH\x00R\x0bupdateEvent\x12>\n\x0c\x64\x65lete_event\x18\x03 \x01(\x0b\x32\x19.cloudidl.app.DeleteEventH\x00R\x0b\x64\x65leteEventB\x07\n\x05\x65vent\"D\n\x13UpdateStatusRequest\x12-\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x03\x61pp\";\n\x14UpdateStatusResponse\x12#\n\x03\x61pp\x18\x01 \x01(\x0b\x32\x11.cloudidl.app.AppR\x03\x61pp\"L\n\x0cLeaseRequest\x12<\n\x02id\x18\x01 \x01(\x0b\x32\".cloudidl.common.ClusterIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\"6\n\rLeaseResponse\x12%\n\x04\x61pps\x18\x01 \x03(\x0b\x32\x11.cloudidl.app.AppR\x04\x61ppsB\x9e\x01\n\x10\x63om.cloudidl.appB\x0f\x41ppPayloadProtoH\x02P\x01Z&github.com/unionai/cloud/gen/pb-go/app\xa2\x02\x03\x43\x41X\xaa\x02\x0c\x43loudidl.App\xca\x02\x0c\x43loudidl\\App\xe2\x02\x18\x43loudidl\\App\\GPBMetadata\xea\x02\rCloudidl::Appb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'app.app_payload_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\020com.cloudidl.appB\017AppPayloadProtoH\002P\001Z&github.com/unionai/cloud/gen/pb-go/app\242\002\003CAX\252\002\014Cloudidl.App\312\002\014Cloudidl\\App\342\002\030Cloudidl\\App\\GPBMetadata\352\002\rCloudidl::App'
  _CREATEREQUEST.fields_by_name['app']._options = None
  _CREATEREQUEST.fields_by_name['app']._serialized_options = b'\372B\005\212\001\002\020\001'
  _GETREQUEST.oneofs_by_name['identifier']._options = None
  _GETREQUEST.oneofs_by_name['identifier']._serialized_options = b'\370B\001'
  _GETREQUEST.fields_by_name['app_id']._options = None
  _GETREQUEST.fields_by_name['app_id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _GETREQUEST.fields_by_name['ingress']._options = None
  _GETREQUEST.fields_by_name['ingress']._serialized_options = b'\372B\005\212\001\002\020\001'
  _UPDATEREQUEST.fields_by_name['app']._options = None
  _UPDATEREQUEST.fields_by_name['app']._serialized_options = b'\372B\005\212\001\002\020\001'
  _DELETEREQUEST.fields_by_name['app_id']._options = None
  _DELETEREQUEST.fields_by_name['app_id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _LISTREQUEST.oneofs_by_name['filter_by']._options = None
  _LISTREQUEST.oneofs_by_name['filter_by']._serialized_options = b'\370B\001'
  _LISTREQUEST.fields_by_name['org']._options = None
  _LISTREQUEST.fields_by_name['org']._serialized_options = b'\372B\004r\002\020\001'
  _LISTREQUEST.fields_by_name['cluster_id']._options = None
  _LISTREQUEST.fields_by_name['cluster_id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _LISTREQUEST.fields_by_name['project']._options = None
  _LISTREQUEST.fields_by_name['project']._serialized_options = b'\372B\005\212\001\002\020\001'
  _WATCHREQUEST.oneofs_by_name['target']._options = None
  _WATCHREQUEST.oneofs_by_name['target']._serialized_options = b'\370B\001'
  _WATCHREQUEST.fields_by_name['org']._options = None
  _WATCHREQUEST.fields_by_name['org']._serialized_options = b'\372B\004r\002\020\001'
  _WATCHREQUEST.fields_by_name['cluster_id']._options = None
  _WATCHREQUEST.fields_by_name['cluster_id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _WATCHREQUEST.fields_by_name['project']._options = None
  _WATCHREQUEST.fields_by_name['project']._serialized_options = b'\372B\005\212\001\002\020\001'
  _WATCHREQUEST.fields_by_name['app_id']._options = None
  _WATCHREQUEST.fields_by_name['app_id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _UPDATESTATUSREQUEST.fields_by_name['app']._options = None
  _UPDATESTATUSREQUEST.fields_by_name['app']._serialized_options = b'\372B\005\212\001\002\020\001'
  _LEASEREQUEST.fields_by_name['id']._options = None
  _LEASEREQUEST.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _globals['_CREATEREQUEST']._serialized_start=134
  _globals['_CREATEREQUEST']._serialized_end=196
  _globals['_CREATERESPONSE']._serialized_start=198
  _globals['_CREATERESPONSE']._serialized_end=251
  _globals['_GETREQUEST']._serialized_start=254
  _globals['_GETREQUEST']._serialized_end=407
  _globals['_GETRESPONSE']._serialized_start=409
  _globals['_GETRESPONSE']._serialized_end=459
  _globals['_UPDATEREQUEST']._serialized_start=461
  _globals['_UPDATEREQUEST']._serialized_end=523
  _globals['_UPDATERESPONSE']._serialized_start=525
  _globals['_UPDATERESPONSE']._serialized_end=578
  _globals['_DELETEREQUEST']._serialized_start=580
  _globals['_DELETEREQUEST']._serialized_end=654
  _globals['_DELETERESPONSE']._serialized_start=656
  _globals['_DELETERESPONSE']._serialized_end=672
  _globals['_LISTREQUEST']._serialized_start=675
  _globals['_LISTREQUEST']._serialized_end=944
  _globals['_LISTRESPONSE']._serialized_start=946
  _globals['_LISTRESPONSE']._serialized_end=1021
  _globals['_WATCHREQUEST']._serialized_start=1024
  _globals['_WATCHREQUEST']._serialized_end=1296
  _globals['_CREATEEVENT']._serialized_start=1298
  _globals['_CREATEEVENT']._serialized_end=1348
  _globals['_UPDATEEVENT']._serialized_start=1350
  _globals['_UPDATEEVENT']._serialized_end=1459
  _globals['_DELETEEVENT']._serialized_start=1461
  _globals['_DELETEEVENT']._serialized_end=1511
  _globals['_WATCHRESPONSE']._serialized_start=1514
  _globals['_WATCHRESPONSE']._serialized_end=1730
  _globals['_UPDATESTATUSREQUEST']._serialized_start=1732
  _globals['_UPDATESTATUSREQUEST']._serialized_end=1800
  _globals['_UPDATESTATUSRESPONSE']._serialized_start=1802
  _globals['_UPDATESTATUSRESPONSE']._serialized_end=1861
  _globals['_LEASEREQUEST']._serialized_start=1863
  _globals['_LEASEREQUEST']._serialized_end=1939
  _globals['_LEASERESPONSE']._serialized_start=1941
  _globals['_LEASERESPONSE']._serialized_end=1995
# @@protoc_insertion_point(module_scope)



from click import Option, UsageError


# See https://stackoverflow.com/a/37491504/499285 and https://stackoverflow.com/a/44349292/499285
class MutuallyExclusiveOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            kwargs["help"] = help + f" Mutually exclusive with {', '.join(self.mutually_exclusive)}."
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        self_present = self.name in opts
        others_present = self.mutually_exclusive.intersection(opts)

        if others_present:
            if self_present:
                raise UsageError(
                    f"Illegal usage: {self.name} is mutually exclusive with {', '.join(self.mutually_exclusive)}."
                )
            else:
                self.prompt = None

        return super().handle_parse_result(ctx, opts, args)



import glob
import os
import platform
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

from flyteidl.core.literals_pb2 import KeyValuePair
from flyteidl.core.tasks_pb2 import Container, ContainerPort
from flyteidl.core.tasks_pb2 import Resources as ResourcesIDL
from flytekit import FlyteContextManager, ImageSpec, Resources
from flytekit.core.artifact import ArtifactQuery
from mashumaro.codecs.json import JSONEncoder

from union.app._common import patch_get_flytekit_for_pypi, patch_get_unionai_for_pypi
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.app.app_definition_pb2 import AutoscalingConfig, Identifier, IngressConfig, Meta, Replicas, Spec


@dataclass
class AppRegistry:
    """Keeps track of all user defined applications"""

    apps: Dict[str, "App"] = field(default_factory=dict)

    def add_app(self, app: "App"):
        execution_state = FlyteContextManager.current_context().execution_state
        if execution_state.mode is None and app.name in self.apps:
            # This is running in a non execute context, so we check if the app name is already used:
            msg = f"App with name: {app.name} was already defined, please use a different name"
            raise ValueError(msg)

        self.apps[app.name] = app


APP_REGISTRY = AppRegistry()
# TODO: Add more protocols here when we deploy to another cloud providers
SUPPORTED_FS_PROTOCOLS = ["s3://", "s3a://", "gs://"]


@dataclass
class URLQuery:
    name: str
    public: bool


ENV_NAME_RE = re.compile("^[_a-zA-Z][_a-zA-Z0-9]*$")


def _has_file_extension(file_path) -> bool:
    _, ext = os.path.splitext(file_path)
    return bool(ext)


@dataclass
class Input:
    """
    Input for application.

    :param name: Name of input.
    :param value: Value for input.
    :param env_name: Environment name to set the value in the serving environment.
    :param auto_download: When True, the input will be automatically downloaded. This
        only works if the value refers to an item in a object store. i.e. `s3://...`
    """

    class Type(Enum):
        File = "file"
        Directory = "directory"
        String = "string"

    name: str
    value: Union[str, ArtifactQuery, URLQuery]
    env_name: Optional[str] = None
    type: Optional[Type] = None
    auto_download: bool = False
    dest: Optional[str] = None

    def __post_init__(self):
        if self.env_name is not None and ENV_NAME_RE.match(self.env_name) is None:
            msg = f"env_name ({self.env_name}) is not a valid environment name for shells"
            raise ValueError(msg)


@dataclass
class MaterializedInput:
    value: str
    type: Optional[Input.Type] = None


@dataclass
class AppSerializationSettings:
    """Runtime settings for creating an AppIDL"""

    org: str
    project: str
    domain: str

    desired_state: Spec.DesiredState
    materialized_inputs: dict[str, MaterializedInput] = field(default_factory=dict)
    additional_distribution: Optional[str] = None


@dataclass
class InputBackend:
    """
    Input information for the backend.
    """

    name: str
    value: str
    auto_download: bool
    type: Optional[Input.Type] = None
    env_name: Optional[str] = None
    dest: Optional[str] = None

    def __post_init__(self):
        if self.type is None:
            if any(self.value.startswith(proto) for proto in SUPPORTED_FS_PROTOCOLS):
                if _has_file_extension(self.value):
                    self.type = Input.Type.File
                else:
                    self.type = Input.Type.Directory
            else:
                self.type = Input.Type.String

        if self.type == Input.Type.String:
            # Nothing to download for a string
            self.auto_download = False

    @classmethod
    def from_input(cls, user_input: Input, settings: AppSerializationSettings) -> "InputBackend":
        if isinstance(user_input.value, str):
            value = user_input.value
            input_type = user_input.type
        else:
            # ArtifactQuery or URLQuery
            try:
                materialized_input = settings.materialized_inputs[user_input.name]
                value = materialized_input.value
                input_type = materialized_input.type or user_input.type
            except KeyError:
                msg = f"Did not materialize {user_input.name}"
                raise ValueError(msg)

        return InputBackend(
            name=user_input.name,
            value=value,
            auto_download=user_input.auto_download,
            env_name=user_input.env_name,
            type=input_type,
            dest=user_input.dest,
        )


@dataclass
class ServeConfig:
    """
    Configuration for serve runtime.

    :param code_uri: Location of user code in an object store (s3://...)
    :param user_inputs: User inputs. Passed in by `app.inputs`
    """

    code_uri: str  # location of user code
    inputs: List[InputBackend]


SERVE_CONFIG_ENCODER = JSONEncoder(ServeConfig)


@dataclass
class ResolvedInclude:
    src: str
    dest: str


@dataclass
class App:
    """
    App specification.

    :param name: The name of the application.
    :param container_image: The container image to use for the application.
    :param command: Command to start application.
    :param port: Port application listens to. Currently, this must be 8080 and the application
        must listen on 8080.
    :param requests: Compute resource requests for application.
    :param limits: Compute resource limits for application.
    :param include: Files to include for your application.
    :param inputs: Inputs for the application.
    """

    name: str
    container_image: Union[str, ImageSpec]
    command: List[str]
    port: int
    limits: Resources
    requests: Resources = field(default_factory=Resources)
    min_replicas: int = 0
    max_replicas: int = 1
    include: List[str] = field(default_factory=list)
    inputs: List[Input] = field(default_factory=list)
    env: dict = field(default_factory=dict)

    _include_resolved: List[ResolvedInclude] = field(default_factory=list)

    def __post_init__(self):
        APP_REGISTRY.add_app(self)

    def _resolve_include(self, app_directory: str, cwd: str) -> "App":
        """Resolve include based on the working_dir.

        If a path in `include` is prefixed with "./", then those files are
        assumed to be relative to the file that has the App object.
        """
        relative_prefix = "./"
        seen_dests = set()

        included_resolved = []

        for file in self.include:
            normed_file = os.path.normpath(file)
            if file.startswith(relative_prefix):
                # File is relative to the app_directory:
                src_dir = app_directory
            else:
                src_dir = cwd

            src = os.path.join(src_dir, normed_file)

            if "*" in src:
                new_srcs = glob.glob(src, recursive=True)
            else:
                new_srcs = [src]

            for new_src in new_srcs:
                dest = os.path.relpath(new_src, src_dir)
                if platform.system() == "Windows":
                    # The runtime is linux so the destinations need to be understood by linux.
                    dest = dest.replace(os.sep, "/")

                if dest in seen_dests:
                    msg = f"{dest} is in include multiple times. Please remove one of them."
                    raise ValueError(msg)

                seen_dests.add(dest)

                included_resolved.append(ResolvedInclude(src=new_src, dest=dest))

        self._include_resolved = included_resolved
        return self

    def _get_image(self) -> str:
        if isinstance(self.container_image, str):
            return self.container_image

        # Build for image Spec
        from flytekit.image_spec.image_spec import ImageBuildEngine

        # TODO: Remove this patching
        # Do not install flytekit or union when building image
        with patch_get_flytekit_for_pypi(), patch_get_unionai_for_pypi():
            ImageBuildEngine.build(self.container_image)
        return self.container_image.image_name()

    def _get_command(self, settings: AppSerializationSettings) -> List[str]:
        args = ["union-serve"]

        serve_config = ServeConfig(
            code_uri=settings.additional_distribution,
            inputs=[InputBackend.from_input(user_input, settings) for user_input in self.inputs],
        )
        args.extend(["--config", SERVE_CONFIG_ENCODER.encode(serve_config)])
        return args + ["--"] + self.command

    def _get_resources(self) -> ResourcesIDL:
        requests = []
        limits = []

        if self.requests.cpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.requests.cpu))
            )
        if self.requests.mem is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.requests.mem))
            )
        if self.requests.gpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.requests.gpu))
            )
        if self.requests.ephemeral_storage is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.requests.ephemeral_storage)
                )
            )

        if self.limits.cpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.limits.cpu)))
        if self.limits.mem is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.limits.mem)))
        if self.limits.gpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.limits.gpu)))
        if self.limits.ephemeral_storage is not None:
            limits.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.limits.ephemeral_storage)
                )
            )

        return ResourcesIDL(requests=requests, limits=limits)

    def _get_container(self, settings: AppSerializationSettings) -> Container:
        return Container(
            image=self._get_image(),
            command=[],
            args=self._get_command(settings),
            resources=self._get_resources(),
            ports=[
                ContainerPort(container_port=self.port),
            ],
            env=[KeyValuePair(key=k, value=v) for k, v in self.env.items()],
        )

    def _to_union_idl(self, settings: AppSerializationSettings) -> AppIDL:
        return AppIDL(
            metadata=Meta(
                id=Identifier(
                    org=settings.org,
                    project=settings.project,
                    domain=settings.domain,
                    name=self.name,
                ),
            ),
            spec=Spec(
                container=self._get_container(settings),
                desired_state=settings.desired_state,
                cluster_pool="default",
                ingress=IngressConfig(private=False),
                autoscaling=AutoscalingConfig(
                    replicas=Replicas(
                        min=self.min_replicas,
                        max=self.max_replicas,
                    )
                ),
            ),
        )

    def query_endpoint(self, *, public: bool) -> URLQuery:
        """
        Query for endpoint.

        :param public: Whether to return the public or internal endpoint.
        :returns: Object representing a URL query.
        """
        return URLQuery(name=self.name, public=public)


@dataclass
class Endpoint(App):
    """
    Endpoint specification.

    :param name: The name of the application.
    :param container_image: The container image to use for the application.
    :param command: Command to start application.
    :param port: Port application listens to. Currently, this must be 8080 and the application
        must listen on 8080.
    :param requests: Compute resource requests for application.
    :param limits: Compute resource limits for application.
    :param include: Files to include for your application.
    :param inputs: Inputs for the application.
    """



import importlib.util
import os
import sys
from copy import deepcopy
from pathlib import Path
from textwrap import dedent
from typing import Dict

import rich_click as click

from union.app import App, Input

ENABLE_UNION_SERVING_ENV_VAR = "ENABLE_UNION_SERVING"
ENABLE_UNION_SERVING = os.getenv(ENABLE_UNION_SERVING_ENV_VAR, "0") == "1"


class ApplicationCommand(click.RichCommand):
    def __init__(
        self,
        app: App,
        project: str,
        domain: str,
        func: callable,
        *args,
        **kwargs,
    ):
        self.app = app
        self.project = project
        self.domain = domain
        self.func = func

        kwargs["params"] = [
            click.Option(
                param_decls=[f"--{app_input.name}"],
                required=False,
                type=str,
                help=f"Set {app_input.name}",
            )
            for app_input in app.inputs
        ]

        super().__init__(*args, **kwargs)

    def invoke(self, ctx):
        app = self.app

        # Check if there are any dynamic inputs
        has_new_inputs = False
        new_inputs = []

        for user_input in app.inputs:
            key = user_input.name.lower()
            if key in ctx.params and ctx.params[key]:
                new_inputs.append(
                    Input(
                        name=user_input.name,
                        value=ctx.params[key],
                        auto_download=user_input.auto_download,
                    )
                )
                has_new_inputs = True
            else:
                new_inputs.append(user_input)

        if has_new_inputs:
            app = deepcopy(app)
            app.inputs = new_inputs

        self.func(app=app, project=self.project, domain=self.domain)


class ApplicationForFileGroup(click.RichGroup):
    def __init__(
        self,
        filename: Path,
        func: callable,
        project: str,
        domain: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not filename.exists():
            raise click.ClickException(f"{filename} does not exists")
        self.filename = filename
        self.func = func
        self.project = project
        self.domain = domain

    @property
    def apps(self) -> Dict[str, App]:
        try:
            return self._apps
        except AttributeError:
            module_name = os.path.splitext(os.path.basename(self.filename))[0]
            module_path = os.path.dirname(os.path.abspath(self.filename))

            spec = importlib.util.spec_from_file_location(module_name, self.filename)
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module

            sys.path.append(module_path)
            spec.loader.exec_module(module)

            from union.app._models import APP_REGISTRY

            app_directory = os.path.dirname(os.path.abspath(self.filename))
            cwd = os.getcwd()
            self._apps = {k: v._resolve_include(app_directory, cwd) for k, v in APP_REGISTRY.apps.items()}
            return self._apps

    def list_commands(self, ctx):
        return list(self.apps.keys())

    def get_command(self, ctx, app_name):
        try:
            return ApplicationCommand(
                app=self.apps[app_name],
                name=app_name,
                project=self.project,
                domain=self.domain,
                func=self.func,
            )
        except KeyError:
            valid_names = list(self.apps.keys())
            err_msg = dedent(f"""\
                '{app_name}' is not a valid app name in {self.filename}

                Valid names are: {valid_names}""")
            raise click.ClickException(err_msg)


class ApplicationGroupForFiles(click.RichGroup):
    def __init__(self, func: callable, command_name: str, *args, **kwargs):
        kwargs["params"] = [
            click.Option(
                param_decls=["-p", "--project"],
                required=False,
                type=str,
                default="flytesnacks",
                help=f"Project to {command_name} to",
            ),
            click.Option(
                param_decls=["-d", "--domain"],
                required=False,
                type=str,
                default="development",
                help=f"domain to {command_name} to",
            ),
        ]
        super().__init__(*args, **kwargs)
        self.func = func
        self.command_name = command_name

    @property
    def files(self):
        try:
            return self._files
        except AttributeError:
            self._files = [os.fspath(p) for p in Path(".").glob("*.py") if p.name != "__init__.py"]
            return self._files

    def list_commands(self, ctx):
        return self.files

    def get_command(self, ctx, filename):
        return ApplicationForFileGroup(
            filename=Path(filename),
            name=filename,
            project=ctx.params["project"],
            domain=ctx.params["domain"],
            func=self.func,
            help=f"{self.command_name} application in {filename}",
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: app/app_definition.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from flyteidl.core import security_pb2 as flyteidl_dot_core_dot_security__pb2
from flyteidl.core import tasks_pb2 as flyteidl_dot_core_dot_tasks__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x61pp/app_definition.proto\x12\x0c\x63loudidl.app\x1a\x1c\x66lyteidl/core/security.proto\x1a\x19\x66lyteidl/core/tasks.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x17validate/validate.proto\"\x88\x01\n\nIdentifier\x12\x19\n\x03org\x18\x01 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x03org\x12!\n\x07project\x18\x02 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x07project\x12\x1f\n\x06\x64omain\x18\x03 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x06\x64omain\x12\x1b\n\x04name\x18\x04 \x01(\tB\x07\xfa\x42\x04r\x02\x10\x01R\x04name\"\xd2\x01\n\x04Meta\x12\x32\n\x02id\x18\x01 \x01(\x0b\x32\x18.cloudidl.app.IdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\x12#\n\x08revision\x18\x02 \x01(\x04\x42\x07\xfa\x42\x04\x32\x02(\x00R\x08revision\x12\x36\n\x06labels\x18\x03 \x03(\x0b\x32\x1e.cloudidl.app.Meta.LabelsEntryR\x06labels\x1a\x39\n\x0bLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\x95\x01\n\x03\x41pp\x12\x38\n\x08metadata\x18\x01 \x01(\x0b\x32\x12.cloudidl.app.MetaB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x08metadata\x12&\n\x04spec\x18\x02 \x01(\x0b\x32\x12.cloudidl.app.SpecR\x04spec\x12,\n\x06status\x18\x03 \x01(\x0b\x32\x14.cloudidl.app.StatusR\x06status\"\xec\x01\n\tCondition\x12L\n\x14last_transition_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x12lastTransitionTime\x12R\n\x11\x64\x65ployment_status\x18\x02 \x01(\x0e\x32%.cloudidl.app.Status.DeploymentStatusR\x10\x64\x65ploymentStatus\x12\x18\n\x07message\x18\x03 \x01(\tR\x07message\x12#\n\x08revision\x18\x04 \x01(\x04\x42\x07\xfa\x42\x04\x32\x02(\x00R\x08revision\"\xc5\x04\n\x06Status\x12)\n\x10\x61ssigned_cluster\x18\x01 \x01(\tR\x0f\x61ssignedCluster\x12\x32\n\x10\x63urrent_replicas\x18\x02 \x01(\rB\x07\xfa\x42\x04*\x02(\x00R\x0f\x63urrentReplicas\x12/\n\x07ingress\x18\x03 \x01(\x0b\x32\x15.cloudidl.app.IngressR\x07ingress\x12\x39\n\ncreated_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x42\n\x0flast_updated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rlastUpdatedAt\x12\x37\n\nconditions\x18\x06 \x03(\x0b\x32\x17.cloudidl.app.ConditionR\nconditions\"\xf2\x01\n\x10\x44\x65ploymentStatus\x12!\n\x1d\x44\x45PLOYMENT_STATUS_UNSPECIFIED\x10\x00\x12 \n\x1c\x44\x45PLOYMENT_STATUS_UNASSIGNED\x10\x01\x12\x1e\n\x1a\x44\x45PLOYMENT_STATUS_ASSIGNED\x10\x02\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_PENDING\x10\x03\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_STOPPED\x10\x04\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_STARTED\x10\x05\x12\x1c\n\x18\x44\x45PLOYMENT_STATUS_FAILED\x10\x06\"\x85\x01\n\x07Ingress\x12*\n\npublic_url\x18\x01 \x01(\tB\x0b\xfa\x42\x08r\x06\x88\x01\x01\xd0\x01\x01R\tpublicUrl\x12(\n\tcname_url\x18\x02 \x01(\tB\x0b\xfa\x42\x08r\x06\x88\x01\x01\xd0\x01\x01R\x08\x63nameUrl\x12$\n\x07vpc_url\x18\x03 \x01(\tB\x0b\xfa\x42\x08r\x06\x88\x01\x01\xd0\x01\x01R\x06vpcUrl\"\xda\x04\n\x04Spec\x12\x42\n\tcontainer\x18\x01 \x01(\x0b\x32\x18.flyteidl.core.ContainerB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\tcontainer\x12\x33\n\x03pod\x18\x02 \x01(\x0b\x32\x15.flyteidl.core.K8sPodB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x03pod\x12\x41\n\x0b\x61utoscaling\x18\x03 \x01(\x0b\x32\x1f.cloudidl.app.AutoscalingConfigR\x0b\x61utoscaling\x12\x35\n\x07ingress\x18\x04 \x01(\x0b\x32\x1b.cloudidl.app.IngressConfigR\x07ingress\x12\x44\n\rdesired_state\x18\x05 \x01(\x0e\x32\x1f.cloudidl.app.Spec.DesiredStateR\x0c\x64\x65siredState\x12!\n\x0c\x63luster_pool\x18\x06 \x01(\tR\x0b\x63lusterPool\x12\x32\n\x06images\x18\x07 \x01(\x0b\x32\x1a.cloudidl.app.ImageSpecSetR\x06images\x12I\n\x10security_context\x18\x08 \x01(\x0b\x32\x1e.flyteidl.core.SecurityContextR\x0fsecurityContext\"c\n\x0c\x44\x65siredState\x12\x1d\n\x19\x44\x45SIRED_STATE_UNSPECIFIED\x10\x00\x12\x19\n\x15\x44\x45SIRED_STATE_STOPPED\x10\x01\x12\x19\n\x15\x44\x45SIRED_STATE_STARTED\x10\x02\x42\x12\n\x0b\x61pp_payload\x12\x03\xf8\x42\x01\"A\n\tImageSpec\x12\x10\n\x03tag\x18\x01 \x01(\tR\x03tag\x12\"\n\rbuild_job_url\x18\x02 \x01(\tR\x0b\x62uildJobUrl\"?\n\x0cImageSpecSet\x12/\n\x06images\x18\x01 \x03(\x0b\x32\x17.cloudidl.app.ImageSpecR\x06images\"]\n\rIngressConfig\x12\x18\n\x07private\x18\x01 \x01(\x08R\x07private\x12\x1c\n\tsubdomain\x18\x02 \x01(\tR\tsubdomain\x12\x14\n\x05\x63name\x18\x03 \x01(\tR\x05\x63name\"\xd1\x01\n\x11\x41utoscalingConfig\x12\x32\n\x08replicas\x18\x01 \x01(\x0b\x32\x16.cloudidl.app.ReplicasR\x08replicas\x12\x44\n\x10scaledown_period\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0fscaledownPeriod\x12\x42\n\x0escaling_metric\x18\x03 \x01(\x0b\x32\x1b.cloudidl.app.ScalingMetricR\rscalingMetric\"\x8a\x01\n\rScalingMetric\x12<\n\x0crequest_rate\x18\x01 \x01(\x0b\x32\x19.cloudidl.app.RequestRateR\x0brequestRate\x12;\n\x0b\x63oncurrency\x18\x02 \x01(\x0b\x32\x19.cloudidl.app.ConcurrencyR\x0b\x63oncurrency\"9\n\x0b\x43oncurrency\x12*\n\x0ctarget_value\x18\x01 \x01(\rB\x07\xfa\x42\x04*\x02 \x00R\x0btargetValue\"9\n\x0bRequestRate\x12*\n\x0ctarget_value\x18\x01 \x01(\rB\x07\xfa\x42\x04*\x02(\x00R\x0btargetValue\"@\n\x08Replicas\x12\x19\n\x03min\x18\x01 \x01(\rB\x07\xfa\x42\x04*\x02(\x00R\x03min\x12\x19\n\x03max\x18\x02 \x01(\rB\x07\xfa\x42\x04*\x02(\x00R\x03maxB\xa1\x01\n\x10\x63om.cloudidl.appB\x12\x41ppDefinitionProtoH\x02P\x01Z&github.com/unionai/cloud/gen/pb-go/app\xa2\x02\x03\x43\x41X\xaa\x02\x0c\x43loudidl.App\xca\x02\x0c\x43loudidl\\App\xe2\x02\x18\x43loudidl\\App\\GPBMetadata\xea\x02\rCloudidl::Appb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'app.app_definition_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\020com.cloudidl.appB\022AppDefinitionProtoH\002P\001Z&github.com/unionai/cloud/gen/pb-go/app\242\002\003CAX\252\002\014Cloudidl.App\312\002\014Cloudidl\\App\342\002\030Cloudidl\\App\\GPBMetadata\352\002\rCloudidl::App'
  _IDENTIFIER.fields_by_name['org']._options = None
  _IDENTIFIER.fields_by_name['org']._serialized_options = b'\372B\004r\002\020\001'
  _IDENTIFIER.fields_by_name['project']._options = None
  _IDENTIFIER.fields_by_name['project']._serialized_options = b'\372B\004r\002\020\001'
  _IDENTIFIER.fields_by_name['domain']._options = None
  _IDENTIFIER.fields_by_name['domain']._serialized_options = b'\372B\004r\002\020\001'
  _IDENTIFIER.fields_by_name['name']._options = None
  _IDENTIFIER.fields_by_name['name']._serialized_options = b'\372B\004r\002\020\001'
  _META_LABELSENTRY._options = None
  _META_LABELSENTRY._serialized_options = b'8\001'
  _META.fields_by_name['id']._options = None
  _META.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _META.fields_by_name['revision']._options = None
  _META.fields_by_name['revision']._serialized_options = b'\372B\0042\002(\000'
  _APP.fields_by_name['metadata']._options = None
  _APP.fields_by_name['metadata']._serialized_options = b'\372B\005\212\001\002\020\001'
  _CONDITION.fields_by_name['revision']._options = None
  _CONDITION.fields_by_name['revision']._serialized_options = b'\372B\0042\002(\000'
  _STATUS.fields_by_name['current_replicas']._options = None
  _STATUS.fields_by_name['current_replicas']._serialized_options = b'\372B\004*\002(\000'
  _INGRESS.fields_by_name['public_url']._options = None
  _INGRESS.fields_by_name['public_url']._serialized_options = b'\372B\010r\006\210\001\001\320\001\001'
  _INGRESS.fields_by_name['cname_url']._options = None
  _INGRESS.fields_by_name['cname_url']._serialized_options = b'\372B\010r\006\210\001\001\320\001\001'
  _INGRESS.fields_by_name['vpc_url']._options = None
  _INGRESS.fields_by_name['vpc_url']._serialized_options = b'\372B\010r\006\210\001\001\320\001\001'
  _SPEC.oneofs_by_name['app_payload']._options = None
  _SPEC.oneofs_by_name['app_payload']._serialized_options = b'\370B\001'
  _SPEC.fields_by_name['container']._options = None
  _SPEC.fields_by_name['container']._serialized_options = b'\372B\005\212\001\002\020\001'
  _SPEC.fields_by_name['pod']._options = None
  _SPEC.fields_by_name['pod']._serialized_options = b'\372B\005\212\001\002\020\001'
  _CONCURRENCY.fields_by_name['target_value']._options = None
  _CONCURRENCY.fields_by_name['target_value']._serialized_options = b'\372B\004*\002 \000'
  _REQUESTRATE.fields_by_name['target_value']._options = None
  _REQUESTRATE.fields_by_name['target_value']._serialized_options = b'\372B\004*\002(\000'
  _REPLICAS.fields_by_name['min']._options = None
  _REPLICAS.fields_by_name['min']._serialized_options = b'\372B\004*\002(\000'
  _REPLICAS.fields_by_name['max']._options = None
  _REPLICAS.fields_by_name['max']._serialized_options = b'\372B\004*\002(\000'
  _globals['_IDENTIFIER']._serialized_start=190
  _globals['_IDENTIFIER']._serialized_end=326
  _globals['_META']._serialized_start=329
  _globals['_META']._serialized_end=539
  _globals['_META_LABELSENTRY']._serialized_start=482
  _globals['_META_LABELSENTRY']._serialized_end=539
  _globals['_APP']._serialized_start=542
  _globals['_APP']._serialized_end=691
  _globals['_CONDITION']._serialized_start=694
  _globals['_CONDITION']._serialized_end=930
  _globals['_STATUS']._serialized_start=933
  _globals['_STATUS']._serialized_end=1514
  _globals['_STATUS_DEPLOYMENTSTATUS']._serialized_start=1272
  _globals['_STATUS_DEPLOYMENTSTATUS']._serialized_end=1514
  _globals['_INGRESS']._serialized_start=1517
  _globals['_INGRESS']._serialized_end=1650
  _globals['_SPEC']._serialized_start=1653
  _globals['_SPEC']._serialized_end=2255
  _globals['_SPEC_DESIREDSTATE']._serialized_start=2136
  _globals['_SPEC_DESIREDSTATE']._serialized_end=2235
  _globals['_IMAGESPEC']._serialized_start=2257
  _globals['_IMAGESPEC']._serialized_end=2322
  _globals['_IMAGESPECSET']._serialized_start=2324
  _globals['_IMAGESPECSET']._serialized_end=2387
  _globals['_INGRESSCONFIG']._serialized_start=2389
  _globals['_INGRESSCONFIG']._serialized_end=2482
  _globals['_AUTOSCALINGCONFIG']._serialized_start=2485
  _globals['_AUTOSCALINGCONFIG']._serialized_end=2694
  _globals['_SCALINGMETRIC']._serialized_start=2697
  _globals['_SCALINGMETRIC']._serialized_end=2835
  _globals['_CONCURRENCY']._serialized_start=2837
  _globals['_CONCURRENCY']._serialized_end=2894
  _globals['_REQUESTRATE']._serialized_start=2896
  _globals['_REQUESTRATE']._serialized_end=2953
  _globals['_REPLICAS']._serialized_start=2955
  _globals['_REPLICAS']._serialized_end=3019
# @@protoc_insertion_point(module_scope)



import rich_click as click

from union.app import App
from union.cli._app_common import ApplicationGroupForFiles
from union.remote._app_remote import AppRemote


@click.group(name="deploy")
def deploy():
    """Deploy a resource."""


def deploy_application(app: App, project: str, domain: str):
    app_remote = AppRemote(project=project, domain=domain)
    app_remote.create_or_update(app)


app_help = """Deploy application on Union."""
app_group = ApplicationGroupForFiles(
    name="apps",
    help=app_help,
    func=deploy_application,
    command_name="deploy",
)
deploy.add_command(app_group)



import logging
import os
import re
import time
from pathlib import Path
from typing import Optional

import rich_click as click
from flytekit import Labels, WorkflowExecutionPhase
from flytekit.configuration import (
    AuthType,
    Config,
    ImageConfig,
    PlatformConfig,
    SerializationSettings,
)
from flytekit.models.core.execution import TaskExecutionPhase
from flytekit.models.filters import ValueIn
from flytekit.remote import FlyteTask, FlyteWorkflowExecution
from flytekit.remote.remote import MOST_RECENT_FIRST
from flytekit.tools.translator import Options
from grpc import RpcError
from rich.console import Console

from union._config import (
    AppClientCredentials,
    _encode_app_client_credentials,
    _get_config_obj,
    _get_endpoint_for_login,
    _get_user_handle,
    _is_serverless_endpoint,
    _write_config_to_path,
)
from union._utils import _sanitize_label
from union.cli._common import _get_channel_with_org
from union.cli._option import MutuallyExclusiveOption
from union.internal.identity.app_payload_pb2 import CreateAppRequest
from union.internal.identity.app_service_pb2_grpc import AppsServiceStub
from union.internal.identity.enums_pb2 import (
    ConsentMethod,
    GrantTypes,
    ResponseTypes,
    TokenEndpointAuthMethod,
)
from union.internal.secret.definition_pb2 import SecretIdentifier, SecretSpec
from union.internal.secret.payload_pb2 import CreateSecretRequest
from union.internal.secret.secret_pb2_grpc import SecretServiceStub
from union.remote import UnionRemote
from union.workspace._vscode import (
    _DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE,
    _DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC,
    WorkspaceConfig,
    resolver,
)
from union.workspace._vscode_remote import VSCODE_DEBUGGER_LOG_NAME, _generate_workspace_name, _get_workspace_executions

UNION_WORKSPACE_LABEL_KEY = "union_workspace"
ENABLE_WORKSPACE_DEFINITION_REGISTRATION = os.getenv("ENABLE_WORKSPACE_DEFINITION_REGISTRATION", "0") == "1"

logger = logging.getLogger("union.cli._create")


@click.group()
def create():
    """Create a resource."""


@create.command()
@click.pass_context
@click.argument("name")
@click.option(
    "--value",
    help="Secret value",
    prompt="Enter secret value",
    hide_input=True,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value_file"],
)
@click.option(
    "-f",
    "--value-file",
    help="Path to file containing the secret",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=True),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value"],
)
@click.option("--project", help="Project name")
@click.option("--domain", help="Domain name")
def secret(
    ctx: click.Context,
    name: str,
    value: str,
    value_file: str,
    project: Optional[str],
    domain: Optional[str],
):
    """Create a secret with NAME."""
    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    if value_file:
        with open(value_file, "rb") as f:
            secret_spec = SecretSpec(binary_value=f.read())
    else:
        secret_spec = SecretSpec(string_value=value)

    stub = SecretServiceStub(channel)
    request = CreateSecretRequest(
        id=SecretIdentifier(name=name, domain=domain, project=project, organization=org),
        secret_spec=secret_spec,
    )
    try:
        stub.CreateSecret(request)
        click.echo(f"Created secret with name: {name}")
    except RpcError as e:
        raise click.ClickException(f"Unable to create secret with name: {name}\n{e}") from e


@create.group("api-key")
def api_key():
    """Manage API keys."""


@api_key.command("admin")
@click.option("--name", type=str, help="Name for API key", required=True)
def admin(name: str):
    """Create an api key."""
    platform_obj = _get_config_obj().platform
    normalized_client_name = re.sub("[^0-9a-zA-Z]+", "-", name.lower())
    if _is_serverless_endpoint(platform_obj.endpoint):
        userhandle = _get_user_handle(platform_obj)
        tenant = platform_obj.endpoint.split(".")[0]
        client_id = f"{tenant}-{userhandle}-{normalized_client_name}"
    else:
        client_id = normalized_client_name
    channel, org = _get_channel_with_org(platform_obj)

    stub = AppsServiceStub(channel)
    request = CreateAppRequest(
        organization=org,
        client_id=client_id,
        client_name=client_id,
        grant_types=[GrantTypes.CLIENT_CREDENTIALS, GrantTypes.AUTHORIZATION_CODE],
        redirect_uris=["http://localhost:8080/authorization-code/callback"],
        response_types=[ResponseTypes.CODE],
        token_endpoint_auth_method=TokenEndpointAuthMethod.CLIENT_SECRET_BASIC,
        consent_method=ConsentMethod.CONSENT_METHOD_REQUIRED,
    )

    try:
        response = stub.Create(request)
    except RpcError as e:
        raise click.ClickException(f"Unable to create api-key with name: {name}\n{e}") from e

    click.echo(f"Client ID: {response.app.client_id}")
    click.echo("The following API key will only be shown once. Be sure to keep it safe!")
    click.echo("Configure your headless CLI by setting the following environment variable:")
    click.echo()

    union_api_key = _encode_app_client_credentials(
        AppClientCredentials(
            endpoint=platform_obj.endpoint,
            client_id=response.app.client_id,
            client_secret=response.app.client_secret,
            org=org,
        )
    )
    click.echo(f'export UNION_API_KEY="{union_api_key}"')


@create.command()
@click.option(
    "--auth",
    type=click.Choice(["device-flow", "pkce"]),
    default="pkce",
    help="Authorization method to ues",
)
@click.option(
    "--host", default=None, help="Host to connect to.", cls=MutuallyExclusiveOption, mutually_exclusive=["serverless"]
)
@click.option(
    "--serverless",
    default=False,
    is_flag=True,
    help="Connect to serverless.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["host"],
)
def login(auth: str, host: Optional[str], serverless: bool):
    """Log into Union

    - On Union Serverless run: `union create login --serverless`

    - On Union BYOC run: `union create login --host UNION_TENANT`
    """
    endpoint = _get_endpoint_for_login(host=host, serverless=serverless)

    if auth == "pkce":
        auth_mode = AuthType.PKCE
    else:
        auth_mode = AuthType.DEVICEFLOW

    config = Config.auto().with_params(
        platform=PlatformConfig(
            endpoint=endpoint,
            insecure=False,
            auth_mode=auth_mode,
        )
    )

    try:
        _write_config_to_path(endpoint, auth_mode.value)

        # Accessing the client will trigger authentication
        remote = UnionRemote(config=config)
        remote.client

        if serverless:
            click.echo("Login successful into serverless")
        else:
            click.echo(f"Login successful into {endpoint}")

    except Exception as e:
        raise click.ClickException(f"Unable to login.\n{e}") from e


@create.command()
@click.pass_context
@click.argument("config_file", type=click.Path(path_type=Path))
def workspace(ctx: click.Context, config_file: Path):
    """Create workspace."""
    config_content = config_file.read_text()
    workspace_config = WorkspaceConfig.from_yaml(config_content)

    remote = UnionRemote(default_domain=workspace_config.domain, default_project=workspace_config.project)
    logger.debug(f"Workspace config: {workspace_config}")
    logger.debug(f"Remote: {remote.__dict__}")
    _create_workspace(remote, workspace_config, config_content)


@click.command()
@click.pass_context
@click.argument("config_file", type=click.Path(path_type=Path))
def workspace_definition(ctx: click.Context, config_file: Path):
    """Create a workspace task definition."""
    config_content = config_file.read_text()
    workspace_config = WorkspaceConfig.from_yaml(config_content)

    remote = UnionRemote(default_domain=workspace_config.domain, default_project=workspace_config.project)
    logger.debug(f"Workspace config: {workspace_config}")
    logger.debug(f"Remote: {remote.__dict__}")

    remote_task = _create_workspace_definition(remote, workspace_config)

    console_url = remote.generate_console_url(remote_task)
    console = Console()
    console.print(f"🚀 Created workspace definition {console_url}")


if ENABLE_WORKSPACE_DEFINITION_REGISTRATION:
    create.add_command(workspace_definition)


@create.command("workspace-config")
@click.argument("config_file", type=click.Path(path_type=Path))
@click.option("--init", type=click.Choice(["base_image"]), required=True)
def workspace_config(config_file: Path, init: str):
    """Create workspace config at CONFIG_FILE."""
    if config_file.exists():
        raise click.ClickException(f"{config_file} already exists")

    click.echo(f"Writing workspace configuration to {config_file}")
    if init == "base_image":
        config_file.write_text(_DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE)
    else:
        config_file.write_text(_DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC)


def _create_workspace_definition(remote: UnionRemote, ws_config: WorkspaceConfig) -> FlyteTask:
    name = ws_config.name
    project = ws_config.project
    domain = ws_config.domain

    task = resolver.get_task_for_workspace(ws_config)
    console = Console()
    console.print("✨ Creating workspace definition...")

    try:
        source_root = task.container_image.source_root
    except AttributeError:
        source_root = None

    return remote.register_task(
        task,
        serialization_settings=SerializationSettings(
            project=project, domain=domain, image_config=ImageConfig.auto_default_image(), source_root=source_root
        ),
        version=_generate_workspace_name(name),
    )


def _get_previous_workspace_state_dir(
    remote: UnionRemote,
    ws_config: WorkspaceConfig,
) -> Optional[str]:
    executions, _ = remote.client.list_executions_paginated(
        project=ws_config.project,
        domain=ws_config.domain,
        limit=1,
        filters=[
            ValueIn("execution_tag.key", [UNION_WORKSPACE_LABEL_KEY]),
            ValueIn("execution_tag.value", [_sanitize_label(ws_config.name)]),
            ValueIn("phase", ["SUCCEEDED"]),
        ],
        sort_by=MOST_RECENT_FIRST,
    )
    if len(executions) == 0:
        return None

    ex = remote.sync_execution(
        FlyteWorkflowExecution.promote_from_model(executions[0]),
    )
    return ex.outputs.literals["state_dir"].scalar.blob.uri


def _create_workspace(remote: UnionRemote, ws_config: WorkspaceConfig, config_content: str):
    project = ws_config.project
    domain = ws_config.domain
    name = ws_config.name

    executions = _get_workspace_executions(remote, project, domain)

    for _, ex_name in executions:
        if ex_name == name:
            raise click.ClickException(f"Workspace with name: {name} already exists.")

    remote_task = _create_workspace_definition(remote=remote, ws_config=ws_config)

    prev_state_dir = _get_previous_workspace_state_dir(remote, ws_config)
    console_http_domain = remote.generate_console_http_domain()

    execution = remote.execute(
        remote_task,
        inputs={
            "config_content": config_content,
            "host": remote.config.platform.endpoint,
            "console_http_domain": console_http_domain,
            "state_dir": prev_state_dir,
        },
        options=Options(
            labels=Labels(values={UNION_WORKSPACE_LABEL_KEY: _sanitize_label(ws_config.name)}),
        ),
        project=project,
        domain=domain,
    )

    vscode_count = 0
    console = Console()
    with console.status("✨ Starting workspace...") as status:
        status.update(status="🚀 Starting workspace...")
        while True:
            time.sleep(5)
            execution = remote.sync_execution(execution, sync_nodes=True)
            if execution.is_done:
                url = remote.generate_console_url(execution)
                raise click.ClickException(f"Workspace failed to launch. See {url} for details.")

            if execution.closure.phase != WorkflowExecutionPhase.RUNNING:
                continue

            status.update(status="⏳ Waiting for node to spin up...")
            if not execution.node_executions:
                continue

            all_keys = [e for e in execution.node_executions if e != "start-node"]
            if not all_keys:
                continue

            vscode_key = all_keys[0]
            vscode_node = execution.node_executions[vscode_key]

            if vscode_node.closure.phase != TaskExecutionPhase.RUNNING or not vscode_node.task_executions:
                continue

            task_execution = vscode_node.task_executions[-1]
            if task_execution.closure.phase != TaskExecutionPhase.RUNNING:
                continue

            for log in task_execution.closure.logs:
                if VSCODE_DEBUGGER_LOG_NAME in log.name:
                    status.update(status="⏳ Waiting for VSCode to start...")
                    resolve_uri = remote.generate_console_http_domain() + log.uri
                    vscode_count += 1
                    if vscode_count == 3:
                        # Make sure execution did not fail before showing dynamic log link
                        execution = remote.sync_execution(execution, sync_nodes=True)
                        if execution.is_done:
                            url = remote.generate_console_url(execution)
                            raise click.ClickException(f"Workspace failed to launch. See {url} for details.")

                        click.echo(f"\n🚀 Open Workspace at:\n{resolve_uri}")
                        return



from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class HTTPMethod(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNSPECIFIED: _ClassVar[HTTPMethod]
    GET: _ClassVar[HTTPMethod]
    PUT: _ClassVar[HTTPMethod]
UNSPECIFIED: HTTPMethod
GET: HTTPMethod
PUT: HTTPMethod

class Metadata(_message.Message):
    __slots__ = ["tag"]
    class TagEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    TAG_FIELD_NUMBER: _ClassVar[int]
    tag: _containers.ScalarMap[str, str]
    def __init__(self, tag: _Optional[_Mapping[str, str]] = ...) -> None: ...

class Object(_message.Message):
    __slots__ = ["contents"]
    CONTENTS_FIELD_NUMBER: _ClassVar[int]
    contents: bytes
    def __init__(self, contents: _Optional[bytes] = ...) -> None: ...

class Key(_message.Message):
    __slots__ = ["key"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    key: str
    def __init__(self, key: _Optional[str] = ...) -> None: ...

class PresignPutRequest(_message.Message):
    __slots__ = ["content_md5"]
    CONTENT_MD5_FIELD_NUMBER: _ClassVar[int]
    content_md5: bytes
    def __init__(self, content_md5: _Optional[bytes] = ...) -> None: ...

class PresignGetRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...
